from typing import Any, Final

from zargunlite.model import ZircoliteFieldMappingConfig, ZircoliteFieldMappingSplitConfig

_default_field_mapping_config_data: Final[dict[str, Any]] = {
    "exclusions": ["xmlns"],
    "useless": [None, ""],
    "mappings": {
        "Event.EventData.UserData": "UserData",
        "Event.System.Provider.#attributes.Guid": "Guid",
        "Event.EventData.ContextInfo": "ContextInfo",
        "Event.System.Execution.#attributes.ProcessID": "ProcessID",
        "Event.System.Execution.#attributes.ThreadID": "ThreadID",
        "Event.System.EventID": "EventID",
        "Event.System.EventID.#text": "EventID",
        "Event.System.Channel": "Channel",
        "Event.System.Computer": "Computer",
        "Event.System.Correlation": "Correlation",
        "Event.System.Correlation.#attributes.ActivityID": "ActivityID",
        "Event.System.EventID.#attributes.Qualifiers": "Qualifiers",
        "Event.System.EventRecordID": "EventRecordID",
        "Event.System.Keywords": "Keywords",
        "Event.System.Level": "Level",
        "Event.System.Opcode": "Opcode",
        "Event.System.Provider.#attributes.EventSourceName": "EventSourceName",
        "Event.System.Provider.#attributes.Name": "Provider_Name",
        "Event.System.Security": "Security",
        "Event.System.Security.#attributes.UserID": "UserID",
        "Event.System.Task": "Task",
        "Event.System.TimeCreated.#attributes.SystemTime": "SystemTime",
        "Event.System.Version": "Version",
        "Event.EventData.AccessList": "AccessList",
        "Event.EventData.AccessMask": "AccessMask",
        "Event.EventData.Accesses": "Accesses",
        "Event.EventData.AccountDomain": "AccountDomain",
        "Event.EventData.AccountExpires": "AccountExpires",
        "Event.EventData.AccountName": "AccountName",
        "Event.EventData.AddonName": "AddonName",
        "Event.EventData.Address": "Address",
        "Event.EventData.AddressLength": "AddressLength",
        "Event.EventData.AllowedToDelegateTo": "AllowedToDelegateTo",
        "Event.EventData.Application": "Application",
        "Event.EventData.AttributeLDAPDisplayName": "AttributeLDAPDisplayName",
        "Event.EventData.AttributeValue": "AttributeValue",
        "Event.EventData.AuditPolicyChanges": "AuditPolicyChanges",
        "Event.EventData.AuditSourceName": "AuditSourceName",
        "Event.EventData.AuthenticationPackageName": "AuthenticationPackageName",
        "Event.EventData.Binary": "Binary",
        "Event.EventData.BootMode": "BootMode",
        "Event.EventData.BuildVersion": "BuildVersion",
        "Event.EventData.CallingProcessName": "CallingProcessName",
        "Event.EventData.CallTrace": "CallTrace",
        "Event.EventData.CommandLine": "CommandLine",
        "Event.EventData.Company": "Company",
        "Event.EventData.Context": "Context",
        "Event.EventData.CreationUtcTime": "CreationUtcTime",
        "Event.EventData.CurrentDirectory": "CurrentDirectory",
        "Event.EventData.DCName": "DCName",
        "Event.EventData.Description": "Description",
        "Event.EventData.DestinationAddress": "DestinationAddress",
        "Event.EventData.DestinationHostname": "DestinationHostname",
        "Event.EventData.DestinationIp": "DestinationIp",
        "Event.EventData.DestinationIsIpv6": "DestinationIsIpv6",
        "Event.EventData.DestinationPort": "DestinationPort",
        "Event.EventData.DestinationPortName": "DestinationPortName",
        "Event.EventData.DestPort": "DestPort",
        "Event.EventData.Detail": "Detail",
        "Event.EventData.Details": "Details",
        "Event.EventData.DetectionSource": "DetectionSource",
        "Event.EventData.DeviceClassName": "DeviceClassName",
        "Event.EventData.DeviceDescription": "DeviceDescription",
        "Event.EventData.DeviceName": "DeviceName",
        "Event.EventData.DeviceNameLength": "DeviceNameLength",
        "Event.EventData.DeviceTime": "DeviceTime",
        "Event.EventData.DeviceVersionMajor": "DeviceVersionMajor",
        "Event.EventData.DeviceVersionMinor": "DeviceVersionMinor",
        "Event.EventData.DisplayName": "DisplayName",
        "Event.EventData.EngineVersion": "EngineVersion",
        "Event.EventData.ErrorCode": "ErrorCode",
        "Event.EventData.ErrorDescription": "ErrorDescription",
        "Event.EventData.ErrorMessage": "ErrorMessage",
        "Event.EventData.EventSourceId": "EventSourceId",
        "Event.EventData.EventType": "EventType",
        "Event.EventData.ExtensionId": "ExtensionId",
        "Event.EventData.ExtensionName": "ExtensionName",
        "Event.EventData.ExtraInfo": "ExtraInfo",
        "Event.EventData.FailureCode": "FailureCode",
        "Event.EventData.FailureReason": "FailureReason",
        "Event.EventData.FileVersion": "FileVersion",
        "Event.EventData.FilterHostProcessID": "FilterHostProcessID",
        "Event.EventData.FinalStatus": "FinalStatus",
        "Event.EventData.GrantedAccess": "GrantedAccess",
        "Event.EventData.Group": "Group",
        "Event.EventData.GroupDomain": "GroupDomain",
        "Event.EventData.GroupName": "GroupName",
        "Event.EventData.GroupSid": "GroupSid",
        "Event.EventData.HandleId": "HandleId",
        "Event.EventData.Hash": "Hash",
        "Event.EventData.Hashes": "Hashes",
        "Event.EventData.HiveName": "HiveName",
        "Event.EventData.HomeDirectory": "HomeDirectory",
        "Event.EventData.HomePath": "HomePath",
        "Event.EventData.HostApplication": "HostApplication",
        "Event.EventData.HostName": "HostName",
        "Event.EventData.HostVersion": "HostVersion",
        "Event.EventData.IdleStateCount": "IdleStateCount",
        "Event.EventData.Image": "Image",
        "Event.EventData.ImageLoaded": "ImageLoaded",
        "Event.EventData.ImagePath": "ImagePath",
        "Event.EventData.Initiated": "Initiated",
        "Event.EventData.IntegrityLevel": "IntegrityLevel",
        "Event.EventData.IpAddress": "IpAddress",
        "Event.EventData.IpPort": "IpPort",
        "Event.EventData.KeyLength": "KeyLength",
        "Event.EventData.LayerRTID": "LayerRTID",
        "Event.EventData.LDAPDisplayName": "LDAPDisplayName",
        "Event.EventData.LmPackageName": "LmPackageName",
        "Event.EventData.LogonGuid": "LogonGuid",
        "Event.EventData.LogonHours": "LogonHours",
        "Event.EventData.LogonId": "LogonId",
        "Event.EventData.LogonProcessName": "LogonProcessName",
        "Event.EventData.LogonType": "LogonType",
        "Event.EventData.MajorVersion": "MajorVersion",
        "Event.EventData.Data.#text": "Message",
        "Event.EventData.MinorVersion": "MinorVersion",
        "Event.EventData.NewName": "NewName",
        "Event.EventData.NewProcessId": "NewProcessId",
        "Event.EventData.NewProcessName": "NewProcessName",
        "Event.EventData.NewState": "NewState",
        "Event.EventData.NewThreadId": "NewThreadId",
        "Event.EventData.NewTime": "NewTime",
        "Event.EventData.NewUacValue": "NewUacValue",
        "Event.EventData.NewValue": "NewValue",
        "Event.EventData.NotificationPackageName": "NotificationPackageName",
        "Event.EventData.Number": "Number",
        "Event.EventData.NumberOfGroupPolicyObjects": "NumberOfGroupPolicyObjects",
        "Event.EventData.ObjectClass": "ObjectClass",
        "Event.EventData.ObjectName": "ObjectName",
        "Event.EventData.ObjectServer": "ObjectServer",
        "Event.EventData.ObjectType": "ObjectType",
        "Event.EventData.ObjectValueName": "ObjectValueName",
        "Event.EventData.OldTime": "OldTime",
        "Event.EventData.OldUacValue": "OldUacValue",
        "Event.EventData.OperationType": "OperationType",
        "Event.EventData.OriginalFileName": "OriginalFileName",
        "Event.EventData.PackageName": "PackageName",
        "Event.EventData.ParentCommandLine": "ParentCommandLine",
        "Event.EventData.ParentImage": "ParentImage",
        "Event.EventData.ParentProcessGuid": "ParentProcessGuid",
        "Event.EventData.ParentProcessId": "ParentProcessId",
        "Event.EventData.PasswordLastSet": "PasswordLastSet",
        "Event.EventData.Payload": "Payload",
        "Event.EventData.PerfStateCount": "PerfStateCount",
        "Event.EventData.PipeName": "PipeName",
        "Event.EventData.PreviousTime": "PreviousTime",
        "Event.EventData.PrimaryGroupId": "PrimaryGroupId",
        "Event.EventData.PrivilegeList": "PrivilegeList",
        "Event.EventData.ProcessCommandLine": "ProcessCommandLine",
        "Event.EventData.ProcessGuid": "ProcessGuid",
        "Event.EventData.ProcessId": "ProcessId",
        "Event.EventData.ProcessName": "ProcessName",
        "Event.EventData.ProcessingMode": "ProcessingMode",
        "Event.EventData.ProcessingTimeInMilliseconds": "ProcessingTimeInMilliseconds",
        "Event.EventData.Product": "Product",
        "Event.EventData.ProfilePath": "ProfilePath",
        "Event.EventData.Properties": "Properties",
        "Event.EventData.Protocol": "Protocol",
        "Event.EventData.ProtocolHostProcessID": "ProtocolHostProcessID",
        "Event.EventData.PuaCount": "PuaCount",
        "Event.EventData.PuaPolicyId": "PuaPolicyId",
        "Event.EventData.Publisher": "Publisher",
        "Event.EventData.QfeVersion": "QfeVersion",
        "Event.EventData.QueryName": "QueryName",
        "Event.EventData.QueryResults": "QueryResults",
        "Event.EventData.QueryStatus": "QueryStatus",
        "Event.EventData.RelativeTargetName": "RelativeTargetName",
        "Event.EventData.ResourceManager": "ResourceManager",
        "Event.EventData.RetryMinutes": "RetryMinutes",
        "Event.EventData.RuleName": "RuleName",
        "Event.EventData.SamAccountName": "SAMAccountName",
        "Event.EventData.SchemaVersion": "SchemaVersion",
        "Event.EventData.ScriptPath": "ScriptPath",
        "Event.EventData.ScriptBlockText": "ScriptBlockText",
        "Event.EventData.SecurityPackageName": "SecurityPackageName",
        "Event.EventData.ServerID": "ServerID",
        "Event.EventData.ServerURL": "ServerURL",
        "Event.EventData.Service": "Service",
        "Event.EventData.ServiceName": "ServiceName",
        "Event.EventData.ServicePrincipalNames": "ServicePrincipalNames",
        "Event.EventData.ServiceType": "ServiceType",
        "Event.EventData.ServiceVersion": "ServiceVersion",
        "Event.EventData.ShareLocalPath": "ShareLocalPath",
        "Event.EventData.ShareName": "ShareName",
        "Event.EventData.ShutdownActionType": "ShutdownActionType",
        "Event.EventData.ShutdownEventCode": "ShutdownEventCode",
        "Event.EventData.ShutdownReason": "ShutdownReason",
        "Event.EventData.SidHistory": "SidHistory",
        "Event.EventData.Signature": "Signature",
        "Event.EventData.SignatureStatus": "SignatureStatus",
        "Event.EventData.Signed": "Signed",
        "Event.EventData.SourceAddress": "SourceAddress",
        "Event.EventData.SourceHostname": "SourceHostname",
        "Event.EventData.SourceImage": "SourceImage",
        "Event.EventData.SourceIp": "SourceIp",
        "Event.EventData.SourceNetworkAddress": "SourceNetworkAddress",
        "Event.EventData.SourceIsIpv6": "SourceIsIpv6",
        "Event.EventData.SourcePort": "SourcePort",
        "Event.EventData.SourcePortName": "SourcePortName",
        "Event.EventData.SourceProcessGuid": "SourceProcessGuid",
        "Event.EventData.SourceProcessId": "SourceProcessId",
        "Event.EventData.StartAddress": "StartAddress",
        "Event.EventData.StartFunction": "StartFunction",
        "Event.EventData.StartModule": "StartModule",
        "Event.EventData.StartTime": "StartTime",
        "Event.EventData.StartType": "StartType",
        "Event.EventData.State": "State",
        "Event.EventData.Status": "Status",
        "Event.EventData.StopTime": "StopTime",
        "Event.EventData.SubStatus": "SubStatus",
        "Event.EventData.SubjectDomainName": "SubjectDomainName",
        "Event.EventData.SubjectLogonId": "SubjectLogonId",
        "Event.EventData.SubjectUserName": "SubjectUserName",
        "Event.EventData.SubjectUserSid": "SubjectUserSid",
        "Event.EventData.TSId": "TSId",
        "Event.EventData.TargetDomainName": "TargetDomainName",
        "Event.EventData.TargetFilename": "TargetFileName",
        "Event.EventData.TargetImage": "TargetImage",
        "Event.EventData.TargetInfo": "TargetInfo",
        "Event.EventData.TargetLogonGuid": "TargetLogonGuid",
        "Event.EventData.TargetLogonId": "TargetLogonId",
        "Event.EventData.TargetObject": "TargetObject",
        "Event.EventData.TargetProcessAddress": "TargetProcessAddress",
        "Event.EventData.TargetProcessGuid": "TargetProcessGuid",
        "Event.EventData.TargetProcessId": "TargetProcessId",
        "Event.EventData.TargetServerName": "TargetServerName",
        "Event.EventData.TargetSid": "TargetSid",
        "Event.EventData.TargetUserName": "TargetUserName",
        "Event.EventData.TargetUserSid": "TargetUserSid",
        "Event.EventData.TaskContent": "TaskContent",
        "Event.EventData.TaskContentNew": "TaskContentNew",
        "Event.EventData.TaskName": "TaskName",
        "Event.EventData.TerminalSessionId": "TerminalSessionId",
        "Event.EventData.ThrottleStateCount": "ThrottleStateCount",
        "Event.EventData.TicketEncryptionType": "TicketEncryptionType",
        "Event.EventData.TicketOptions": "TicketOptions",
        "Event.EventData.TimeSource": "TimeSource",
        "Event.EventData.TokenElevationType": "TokenElevationType",
        "Event.EventData.TransactionId": "TransactionId",
        "Event.EventData.TransmittedServices": "TransmittedServices",
        "Event.EventData.User": "User",
        "Event.EventData.UserAccountControl": "UserAccountControl",
        "Event.EventData.UserParameters": "UserParameters",
        "Event.EventData.UserPrincipalName": "UserPrincipalName",
        "Event.EventData.UserSid": "UserSid",
        "Event.EventData.UserWorkstations": "UserWorkstations",
        "Event.EventData.UtcTime": "UtcTime",
        "Event.EventData.Version": "Version",
        "Event.EventData.Workstation": "Workstation",
        "Event.EventData.WorkstationName": "WorkstationName",
        "Event.EventData.updateGuid": "updateGuid",
        "Event.EventData.updateRevisionNumber": "updateRevisionNumber",
        "Event.EventData.updateTitle": "updateTitle",
        "Event.EventData.ParentIntegrityLevel": "ParentIntegrityLevel",
        "Event.EventData.ParentUser": "ParentUser",
    },
    "alias": {},
    "split": {
        "Hash": {"separator": ",", "equal": "="},
        "Hashes": {"separator": ",", "equal": "="},
        "ConfigurationFileHash": {"separator": ",", "equal": "="},
    },
}

DEFAULT_FIELD_MAPPING_CONFIG: Final = ZircoliteFieldMappingConfig(
    exclusions=_default_field_mapping_config_data["exclusions"],
    useless=_default_field_mapping_config_data["useless"],
    mappings=_default_field_mapping_config_data["mappings"],
    alias=_default_field_mapping_config_data["alias"],
    split={
        k: ZircoliteFieldMappingSplitConfig(separator=v["separator"], equal=v["equal"])
        for k, v in _default_field_mapping_config_data["split"].items()
    },
)

EMPTY_FIELD_MAPPING_CONFIG: Final = ZircoliteFieldMappingConfig()



import os
import re
import sqlite3
import string
from collections.abc import Collection, Iterable, Mapping, Sequence
from contextlib import closing
from types import TracebackType
from typing import Any

from zargunlite.model import ZargunException, ZircoliteRule, ZircoliteRuleMatchResult


def sqlite_udf_regexp(x: str, y: object | None) -> int:
    if y is None:
        return 0
    if re.search(x, str(y)):
        return 1
    else:
        return 0


def strict_field_name_check(field_name: str) -> bool:
    return all(c in string.ascii_letters + string.digits + "_" for c in field_name)


def repr_to_sqlite_name_literal(s: str) -> str:
    # FIXME: escape more special chars
    return "`{}`".format(s.replace("`", "``"))


def repr_to_sqlite_value_literal(v: object) -> str:
    s = str(v)
    if isinstance(v, int):
        return s
    safe_chars = string.digits + string.ascii_letters + string.punctuation + " "
    if all(c in safe_chars for c in s):
        return "'{}'".format(s.replace("'", "''"))
    else:
        return f"X'{s.encode().hex()}'"


class ZargunCore:
    __slots__ = ("_db_location", "_db_connection", "_limit")

    def __init__(
        self,
        *,
        db_location: str | bytes | os.PathLike[str] | os.PathLike[bytes] = ":memory:",
        limit: int = -1,
    ) -> None:
        self._db_location = db_location
        self._db_connection = self._create_connection()
        self._limit = limit

    def _create_connection(self) -> sqlite3.Connection:
        conn = sqlite3.connect(self._db_location)
        conn.row_factory = sqlite3.Row
        conn.create_function("regexp", 2, sqlite_udf_regexp)
        return conn

    def close(self) -> None:
        self._db_connection.close()

    def __enter__(self) -> "ZargunCore":  # TODO: when increase python require to >=3.11, we can use typing.Self
        return self

    def __exit__(
        self, exc_type: type[BaseException] | None, exc_value: BaseException | None, traceback: TracebackType | None
    ) -> None:
        _ = exc_type, exc_value
        self.close()

    def _execute_sql(self, sql: str, params: Mapping[str, object] | Sequence[object] = ()) -> list[sqlite3.Row]:
        with closing(self._db_connection.cursor()) as cursor:
            r = cursor.execute(sql, params)
            return r.fetchall()

    def _create_table(self, field_defs: list[tuple[str, str]]) -> None:
        fields_part = "".join(f"{repr_to_sqlite_name_literal(field)} {typ}, " for field, typ in field_defs)
        stmt = f"CREATE TABLE `logs` ( `row_id` INTEGER, {fields_part} PRIMARY KEY(row_id AUTOINCREMENT) );"
        self._execute_sql(stmt)

    def _insert_data_row(self, d: Mapping[str, Any]) -> None:
        column_define_list = []
        value_list = []
        for k, v in d.items():
            column_define = repr_to_sqlite_name_literal(k)
            column_define_list.append(column_define)
            value_list.append(v)

        columns_define = ", ".join(column_define_list)
        parameters_define = ", ".join("?" * len(value_list))
        if column_define_list:
            insert_stmt = f"INSERT INTO `logs` ({columns_define}) VALUES ({parameters_define});"
        else:
            insert_stmt = "INSERT INTO `logs` DEFAULT VALUES;"
        self._execute_sql(insert_stmt, value_list)

    def load_data(
        self,
        data: Collection[Mapping[str, Any]] | Iterable[Mapping[str, Any]],
        *,
        fields: Collection[tuple[str, type]] | None = None,
        strict_field_name: bool = True,
    ) -> None:
        if not fields:
            assert isinstance(data, Collection)
            field_map: dict[str, type] = {}
            for d in data:  # assert data is Collection because we need extra iter on it to extract fields
                field_map.update({k: type(v) for k, v in d.items()})
        else:
            # here, data can be any iterable object
            field_map = {k: t for k, t in fields}

        if strict_field_name:
            for field_name in field_map.keys():
                if not strict_field_name_check(field_name):
                    raise ZargunException(f"strict field name check failed: {field_name!r}")

        field_name_lower_seens: set[str] = set()
        field_define_list: list[tuple[str, str]] = []
        for field_name, field_type in field_map.items():
            field_name_lower = field_name.lower()
            if field_name_lower in field_name_lower_seens:
                # sqlite is case insensitive, so must drop duplicate one
                continue
            field_name_lower_seens.add(field_name_lower)

            sql_type_define = "INTEGER" if issubclass(field_type, int) else "TEXT COLLATE NOCASE"
            field_define = (field_name, sql_type_define)
            field_define_list.append(field_define)
        self._create_table(field_define_list)

        for d in data:
            self._insert_data_row(d)

    def create_index(self, field: str) -> None:
        sql = "CREATE INDEX {} ON `logs` ({});".format(
            repr_to_sqlite_name_literal(f"idx_{field}"),
            repr_to_sqlite_name_literal(field),
        )
        self._execute_sql(sql)

    def execute_sqlite_query(self, sql: str) -> list[dict[str, Any]]:
        r = []
        try:
            rows = self._execute_sql(sql)
            r = [{k: v for k, v in zip(row.keys(), row) if v is not None} for row in rows]
        except sqlite3.OperationalError as e:
            if "no such column" in str(e):  # sqlite3.OperationalError: no such column: <>
                pass
            else:
                raise
        return r

    def execute_zircolite_rule(self, rule: ZircoliteRule) -> ZircoliteRuleMatchResult:
        all_count = 0
        all_matches = []
        for sql in rule.rule:
            matches = self.execute_sqlite_query(sql)
            count = len(matches)
            if count > 0 and (self._limit < 0 or count <= self._limit):
                all_count += count
                all_matches.extend(matches)
        r = ZircoliteRuleMatchResult(
            title=rule.title,
            id=rule.id,
            description=rule.description,
            sigmafile=rule.filename,
            sigma=list(rule.rule),
            rule_level=rule.level,
            tags=list(rule.tags),
            count=all_count,
            matches=all_matches,
        )
        return r



try:
    from zargunlite._version import __version__
except ImportError:
    from importlib.metadata import PackageNotFoundError, version

    try:
        __version__ = version("zargunlite")
    except PackageNotFoundError:
        pass

__all__ = ["__version__"]



from collections.abc import Generator

import pytest

from zargunlite.core import ZargunCore
from zargunlite.model import ZircoliteRule, ZircoliteRuleMatchResult


@pytest.fixture
def zargun_core_instance() -> Generator[ZargunCore, None, None]:
    with ZargunCore() as instance:
        yield instance


@pytest.fixture
def zargun_core_instance_with_data(zargun_core_instance: ZargunCore) -> ZargunCore:
    data: list[dict[str, object]] = [
        {"category": "c1", "fieldA": "value1"},
        {"category": "c2", "fieldB": 22, "fieldD": 44},
        {"category": "c1", "fieldA": "value111", "fieldB": 222, "fieldC": "value333"},
        {"category": "c3", "fieldD": 44},
        {"category": "c4", "fieldD": 44},
        {"category": "c4", "fieldE": "000/abc/def/abc/ghi"},
        {"category": "c4", "fieldE": "111/abc/def/cba/ghi"},
        {"category": "c5", "fieldF": "JKL"},
        {"category": "c5", "fieldF": "jkl"},
    ]
    zargun_core_instance.load_data(data)
    zargun_core_instance.create_index("category")
    return zargun_core_instance


def test_core_sqlite_exception() -> None:
    with ZargunCore() as instance:
        instance.load_data([{"field": "value"}], fields=[("field", str)])
        with pytest.raises(Exception):
            instance.execute_sqlite_query("THIS IS AN INVALID SQL")


def test_core_sqlite_query_single(zargun_core_instance_with_data: ZargunCore) -> None:
    r = zargun_core_instance_with_data.execute_sqlite_query("SELECT * FROM logs WHERE fieldB = '22'")
    assert r == [{"row_id": 2, "category": "c2", "fieldB": 22, "fieldD": 44}]


def test_core_sqlite_query_single_nocase_string(zargun_core_instance_with_data: ZargunCore) -> None:
    r = zargun_core_instance_with_data.execute_sqlite_query("SELECT * FROM logs WHERE fieldF = 'Jkl'")
    assert r == [{"row_id": 8, "category": "c5", "fieldF": "JKL"}, {"row_id": 9, "category": "c5", "fieldF": "jkl"}]


def test_core_sqlite_query_like(zargun_core_instance_with_data: ZargunCore) -> None:
    r = zargun_core_instance_with_data.execute_sqlite_query("SELECT * FROM logs WHERE fieldA LIKE '%111'")
    assert r == [{"row_id": 3, "category": "c1", "fieldA": "value111", "fieldB": 222, "fieldC": "value333"}]


def test_core_sqlite_query_regex(zargun_core_instance_with_data: ZargunCore) -> None:
    query = r"SELECT * FROM logs WHERE fieldE REGEXP '\d+/([a-z]+)/def/\1/ghi'"
    r = zargun_core_instance_with_data.execute_sqlite_query(query)
    assert r == [{"row_id": 6, "category": "c4", "fieldE": "000/abc/def/abc/ghi"}]


def test_core_sqlite_query_no_such_column(zargun_core_instance_with_data: ZargunCore) -> None:
    r = zargun_core_instance_with_data.execute_sqlite_query("SELECT * FROM logs WHERE nosuchcolumn = ''")
    assert r == []


def test_core_zircolite_rule(zargun_core_instance_with_data: ZargunCore) -> None:
    rule = ZircoliteRule(
        title="Test",
        id="id",
        status="status",
        description="description",
        author="author",
        tags=["tag1"],
        falsepositives=[],
        level="level",
        rule=["SELECT * FROM logs WHERE (category='c2' OR category='c3') AND fieldD=44"],
        filename="filename",
    )
    r = zargun_core_instance_with_data.execute_zircolite_rule(rule)
    assert r == ZircoliteRuleMatchResult(
        title="Test",
        id="id",
        description="description",
        sigmafile="filename",
        sigma=["SELECT * FROM logs WHERE (category='c2' OR category='c3') AND fieldD=44"],
        rule_level="level",
        tags=["tag1"],
        count=2,
        matches=[
            {"row_id": 2, "category": "c2", "fieldB": 22, "fieldD": 44},
            {"row_id": 4, "category": "c3", "fieldD": 44},
        ],
    )



from collections.abc import Generator
from typing import Any

import pytest

from zargunlite.core import ZargunCore
from zargunlite.model import ZargunException


@pytest.fixture
def zargun_core_instance() -> Generator[ZargunCore, None, None]:
    with ZargunCore() as instance:
        yield instance


class TestCoreDataAmount:
    @staticmethod
    def test_zero_count_data(zargun_core_instance: ZargunCore) -> None:
        zargun_core_instance.load_data([])
        r = zargun_core_instance.execute_sqlite_query("SELECT * FROM logs")
        assert r == []

    @staticmethod
    def test_lots_of_data(zargun_core_instance: ZargunCore) -> None:
        data = []
        for i in range(1, 1024):
            data.append({"a": i, "b": str(i)})
        zargun_core_instance.load_data(data)
        r = zargun_core_instance.execute_sqlite_query("SELECT * FROM logs WHERE a = 123")
        assert r == [{"row_id": 123, "a": 123, "b": "123"}]

    @staticmethod
    def test_zero_count_fields(zargun_core_instance: ZargunCore) -> None:
        zargun_core_instance.load_data([{}])
        r = zargun_core_instance.execute_sqlite_query("SELECT * FROM logs")
        assert r == [{"row_id": 1}]

    @staticmethod
    def test_lots_of_fields(zargun_core_instance: ZargunCore) -> None:
        d1 = {f"a{i}": i for i in range(1, 1000)}
        d2 = {f"b{i}": i for i in range(1, 1000)}
        data = [d1, d2]
        zargun_core_instance.load_data(data)
        r = zargun_core_instance.execute_sqlite_query("SELECT * FROM logs WHERE a456 = 456")
        assert len(r) == 1


class TestCoreDataSpecialChar:
    @staticmethod
    def test_valid_field_name_in_strict_mode(zargun_core_instance: ZargunCore) -> None:
        d = {"a": 1, "B": 2, "_": 3, "4": 40, "SELECT": "s"}
        data = [d]
        zargun_core_instance.load_data(data, strict_field_name=True)
        sql = "SELECT * FROM logs WHERE a = 1 AND b = 2 AND _ = 3 AND `4` = 40 AND `select` = 's'"
        r = zargun_core_instance.execute_sqlite_query(sql)
        assert r == [{"row_id": 1, "a": 1, "B": 2, "_": 3, "4": 40, "SELECT": "s"}]

    @staticmethod
    @pytest.mark.parametrize("data", [[{"\x00": 1}], [{"a&b": 1}], [{"c d": 1}]])
    def test_invalid_field_name_in_strict_mode(zargun_core_instance: ZargunCore, data: list[dict[str, Any]]) -> None:
        with pytest.raises(ZargunException):
            zargun_core_instance.load_data(data, strict_field_name=True)

    @staticmethod
    def test_special_name_char_in_non_strict_mode(zargun_core_instance: ZargunCore) -> None:
        d = {"a&b": 11, "c d": 22, "e'`f": 33, "\x01": 44}
        data = [d]
        zargun_core_instance.load_data(data, strict_field_name=False)
        r = zargun_core_instance.execute_sqlite_query("SELECT * FROM logs")
        assert r == [{"row_id": 1, "a&b": 11, "c d": 22, "e'`f": 33, "\x01": 44}]

    @staticmethod
    def test_special_value_char(zargun_core_instance: ZargunCore) -> None:
        utf8str = b"\xF0\x9F\x98\x83".decode()
        d = {"a": "a\0bc", "b": "de'f||g''h", "c": "i j`k``l", "d": "\xFF\xFE", "f": utf8str}
        data = [d]
        zargun_core_instance.load_data(data, strict_field_name=False)
        r = zargun_core_instance.execute_sqlite_query(f"SELECT * FROM logs WHERE f='{utf8str}'")
        assert r == [{"row_id": 1, "a": "a\0bc", "b": "de'f||g''h", "c": "i j`k``l", "d": "\xFF\xFE", "f": utf8str}]



import time
from collections.abc import Mapping, Sequence

from zargunlite.default_config import DEFAULT_FIELD_MAPPING_CONFIG
from zargunlite.model import JSONLiteralType, JSONTypeRO, ZircoliteFieldMappingConfig


def flatten_json(x: JSONTypeRO) -> dict[tuple[str, ...], JSONLiteralType] | JSONLiteralType:
    value: dict[tuple[str, ...], JSONLiteralType] | JSONLiteralType | None = None
    if isinstance(x, Mapping):
        value = {}
        for k, v in x.items():
            r = flatten_json(v)
            if isinstance(r, Mapping):
                for sk, sv in r.items():
                    value[(k, *sk)] = sv
            else:
                value[(k,)] = r
    elif isinstance(x, Sequence):
        value = str(x)
    else:
        value = x
    return value


def mapping_field(
    y: dict[tuple[str, ...], JSONLiteralType],
    *,
    field_mapping_config: ZircoliteFieldMappingConfig = DEFAULT_FIELD_MAPPING_CONFIG,
) -> dict[str, JSONLiteralType]:
    result: dict[str, JSONLiteralType] = {}

    for ks, value in y.items():
        raw_key = ".".join(ks)

        # Applying exclusions. Be careful, the key/value pair is discarded if there is a partial match
        if any(exclusion in raw_key for exclusion in field_mapping_config.exclusions):
            continue

        # Excluding useless values (e.g. "null"). The value must be an exact match.
        if value in field_mapping_config.useless:
            continue

        # Applying field mappings
        if raw_key in field_mapping_config.mappings:
            key = field_mapping_config.mappings[raw_key]
        else:
            key = "".join(c for c in ks[-1] if c.isalnum())

        # Applying field splitting
        for tmp_key in {key, raw_key}:
            split_cfg = field_mapping_config.split.get(tmp_key)
            if not (split_cfg is not None and isinstance(value, str)):
                continue
            value_parts = value.split(split_cfg.separator)
            for value_part in value_parts:
                value_sub_parts = value_part.split(split_cfg.equal)
                if len(value_sub_parts) != 2:
                    continue
                k, v = value_sub_parts
                result[k] = result.get(k, v)

        # Applying key and aliases
        result[key] = result.get(key, value)
        for tmp_key in {key, raw_key}:
            alias = field_mapping_config.alias.get(tmp_key)
            if alias is not None:
                result[alias] = result.get(alias, value)

    return result


def filter_time_field(
    d: Mapping[str, JSONLiteralType],
    *,
    time_field: str = "SystemTime",
    time_after: time.struct_time | None = None,
    time_before: time.struct_time | None = None,
) -> bool:
    if (time_after is not None or time_before is not None) and isinstance(time_str_value := d.get(time_field), str):
        time_value = time.strptime(time_str_value.split(".", 1)[0].replace("Z", ""), "%Y-%m-%dT%H:%M:%S")
        return (time_after is None or time_value > time_after) and (time_before is None or time_value < time_before)
    return True



import dataclasses
from collections.abc import Mapping, Sequence
from typing import Any, TypeAlias

JSONLiteralType: TypeAlias = str | int | float | bool | None
JSONType: TypeAlias = dict[str, "JSONType"] | list["JSONType"] | JSONLiteralType
JSONTypeRO: TypeAlias = Mapping[str, "JSONTypeRO"] | Sequence["JSONTypeRO"] | JSONLiteralType

# TODO: when increase python require to >=3.12, we can use the PEP 695 type alias syntax:
# type JSONLiteralType = str | int | float | bool | None
# type JSONType = dict[str, JSONType] | list[JSONType] | JSONLiteralType
# type JSONTypeRO = Mapping[str, "JSONTypeRO"] | Sequence["JSONTypeRO"] | JSONLiteralType


class ZargunException(Exception):
    pass


@dataclasses.dataclass(slots=True, kw_only=True)
class ZircoliteRule:
    title: str
    id: str = ""
    status: str = ""
    description: str = ""
    author: str = ""
    tags: list[str] = dataclasses.field(default_factory=list)
    falsepositives: list[str] = dataclasses.field(default_factory=list)
    level: str = ""
    rule: list[str]
    filename: str = ""


@dataclasses.dataclass(slots=True, kw_only=True)
class ZircoliteRuleMatchResult:
    title: str
    id: str
    description: str
    sigmafile: str
    sigma: list[str]  # ZircoliteRule.rule
    rule_level: str
    tags: list[str]
    count: int
    matches: list[dict[str, Any]]


# ---


@dataclasses.dataclass(slots=True, kw_only=True)
class ZircoliteFieldMappingSplitConfig:
    separator: str
    equal: str


@dataclasses.dataclass(slots=True, kw_only=True)
class ZircoliteFieldMappingConfig:
    exclusions: list[str] = dataclasses.field(default_factory=list)
    useless: list[Any] = dataclasses.field(default_factory=list)
    mappings: dict[str, str] = dataclasses.field(default_factory=dict)
    alias: dict[str, str] = dataclasses.field(default_factory=dict)
    split: dict[str, ZircoliteFieldMappingSplitConfig] = dataclasses.field(default_factory=dict)
